#include "logger.h"
#include "tokeniser.h"
#include "util.h"
#include <assert.h>
#include <stdlib.h>
#include <token.h>
#include <ast.h>
#include <dynarray.h>
#include <parser.h>

typedef struct parser_state_s {
    program_t prog;
    token_t curr_tok;
    token_t* tokens;
} parser_state_t;

token_t* expect_token_type(parser_state_t* state, token_type_t type) {
    const char* exoected_type_str = get_tok_type_str_dbg(state->curr_tok.type);
    if (dynarray_pop(state->tokens, &state->curr_tok) != 0) {
        log_error(&state->curr_tok.loc, "Invalid word, expected %s, got nothing.", exoected_type_str);
        return NULL;
    }
    if (state->curr_tok.type != type) {
        const char* s1 = get_tok_str_dbg(&state->curr_tok);
        log_error(&state->curr_tok.loc, "Invalid word, expected %s, got %s.", exoected_type_str, s1);
        return NULL;
    }
    return &state->curr_tok;
}

token_t* expect_token_type_ex(parser_state_t* state, token_type_t type, int inner_type) {
    const char* exoected_type_str = get_tok_type_str_dbg(state->curr_tok.type);
    if (dynarray_pop(state->tokens, &state->curr_tok) != 0) {
        log_error(&state->curr_tok.loc, "Invalid word, expected %s, got nothing.", exoected_type_str);
        return NULL;
    }
    // NOTE: all inner types are in a union
    if (state->curr_tok.type != type || state->curr_tok.kw_type != inner_type) {
        const char* s1 = get_tok_str_dbg(&state->curr_tok);
        log_error(&state->curr_tok.loc, "Invalid word, expected %s, got %s.", exoected_type_str, s1);
        return NULL;
    }
    return &state->curr_tok;
}

// Tests token type, and inner type, so like, op type, kw type, etc.
// Does not edit state.curr_tok
token_t* test_token_type_ex(parser_state_t* state, token_type_t tok_type, int gen_type) {
    const char* exoected_type_str = get_tok_type_str_dbg(state->curr_tok.type);
    token_t* token = &state->tokens[dynarray_length(state->tokens - 1)];
    // NOTE: tests inner type in a generic way
    if (token->type != tok_type || ((int)token->kw_type) != gen_type) {
        return NULL;
    }
    return token;
}

token_t* test_token_type(parser_state_t* state, token_type_t tok_type) {
    const char* exoected_type_str = get_tok_type_str_dbg(state->curr_tok.type);
    token_t* token = &state->tokens[dynarray_length(state->tokens - 1)];
    if (token->type != tok_type) {
        return NULL;
    }
    return token;
}

int _parse(parser_state_t* state);

program_t parse(const token_t* tokens) {
    // reverses tokens so its way easier to parse;
    dynarray_reverse((void*)tokens);
    parser_state_t state = {0};
    state.tokens = (token_t*)tokens;
    (void)tokens;
    _parse(&state);
    return state.prog;
}

int _parse(parser_state_t* state) {
    while (dynarray_pop(state->tokens, &state->curr_tok) == 0) {
        if (state->curr_tok.type != TT_KW) {
            const char* dbg_s = get_tok_str_dbg(&state->curr_tok);
            log_error(&state->curr_tok.loc, "Invalid word, expected Keyword, got %s.", dbg_s);
            free((void*)dbg_s);
        }
        switch (state->curr_tok.kw_type) {
        case (KW_INCLUDE): {
            expect_token_type(state, TT_PUSH_STR);
            const char* code = read_to_string(state->curr_tok.str_v);
            if (!code) {
                return 1;
            }
            token_t* tokens = tokenise_string((char*)state->curr_tok.str_v, (char*)code);
            if (!tokens) {
                return 1;
            }
            if (_parse(state) != 0) {
                return 1;
            }
        }; break;
        case (KW_CONST): {
            // TODO: Implement compile time calculation of the value
            token_t* tokens = dynarray_create(token_t);

            while (test_token_type_ex(state, TT_KW, KW_END) == NULL) {
                // clang-format off
                if (
                    test_token_type(state, TT_PUSH_STR) || test_token_type(state, TT_PUSH_CSTR) || 
                    test_token_type(state, TT_PUSH_CHAR) || test_token_type(state, TT_PUSH_INT) || 
                    test_token_type(state, TT_PUSH_FLOAT) || test_token_type(state, TT_PUSH_MEM)
                ) { // clang-format on
                    dynarray_pop(state->tokens, &state->curr_tok);
                    dynarray_push(tokens, state->curr_tok);
                }
            }
            if (!expect_token_type_ex(state, TT_KW, KW_END)) {
                return 1;
            }

        }; break;
        case (KW_MEMORY): {

        }; break;
        case (KW_FN): {

        }; break;
        default:
            assert(true && "TODO: parse all kw's");
        }
    }
}
